# -*- coding: utf-8 -*-
"""
Created on Sun Feb 28 09:12:36 2016

@module:    choice_calcs.py
@name:      Choice Calculations
@author:    Timothy Brathwaite
            José Ángel Martín Baos
@summary:   Contains generic functions necessary for calculating choice
            probabilities and for estimating the choice models.
"""

import numpy as np
import scipy.stats
import scipy.linalg
from scipy.linalg import block_diag
from scipy.sparse import hstack

try:
    # Python 3.x does not natively support xrange
    from past.builtins import xrange
except ImportError:
    pass

# Define the boundary values which are not to be exceeded during computation
min_exponent_val = -700
max_exponent_val = 700

max_comp_value = 1e300
min_comp_value = 1e-300
# The value below was determined since its the smallest value, x, for which
# 1 - x != 1
# min_comp_value = 1e-16


def calc_probabilities(beta,
                       design,
                       alt_IDs,
                       rows_to_obs,
                       rows_to_alts,
                       utility_transform,
                       intercept_params=None,
                       shape_params=None,
                       chosen_row_to_obs=None,
                       return_long_probs=False):
    """
    Parameters
    ----------
    beta : 1D or 2D ndarray.
        All elements should by ints, floats, or longs. If 1D, should have 1
        element for each utility coefficient being estimated (i.e.
        num_features). If 2D, should have 1 column for each set of coefficients
        being used to predict the probabilities of each alternative being
        chosen. There should be one row per index coefficient.
    design : 2D or 3D ndarray.
        There should be one row per observation per available alternative.
        There should be one column per utility coefficient being estimated. All
        elements should be ints, floats, or longs. If `len(design.shape) == 3`,
        then beta MUST be 1D.
    alt_IDs : 1D ndarray.
        All elements should be ints. There should be one row per obervation per
        available alternative for the given observation. Elements denote the
        alternative corresponding to the given row of the design matrix.
    rows_to_obs : 2D ndarray.
        There should be one row per observation per available alternative and
        one column per observation. This matrix maps the rows of the design
        matrix to the unique observations (on the columns).
    rows_to_alts : 2D ndarray.
        There should be one row per observation per available alternative and
        one column per possible alternative. This matrix maps the rows of the
        design matrix to the possible alternatives for this dataset.
    utility_transform : callable.
        Should accept a 1D array of systematic utility values, a 1D array of
        alternative IDs, and miscellaneous args and kwargs. Should return a 1D
        array whose elements contain the appropriately transformed systematic
        utility values, based on the current model being evaluated.
    intercept_params : 1D ndarray, or None, optional.
        If an array, each element should be an int, float, or long. For
        identifiability, there should be J- 1 elements where J is the total
        number of observed alternatives for this dataset. Default == None.
    shape_params : 1D ndarray, or None, optional.
        If an array, each element should be an int, float, or long. There
        should be one value per shape parameter of the model being used.
        Default == None.
    chosen_row_to_obs :  2D scipy sparse array, or None, optional.
        There should be one row per observation per available alternative and
        one column per observation. This matrix indicates, for each observation
        (on the columns), which rows of the design matrix were the realized
        outcome. If an array is passed then an array of shape
        (num_observations,) will be returned and each element will be the
        probability of the realized outcome of the given observation.
        Default == None.
    return_long_probs :  bool, optional.
        Indicates whether or not the long format probabilities (a 1D numpy array
        with one element per observation per available alternative) should be
        returned. Default == False.

    Returns
    -------
    numpy array or tuple of two numpy arrays.
        If `chosen_row_to_obs` is passed AND `return_long_probs is True`, then
        the tuple `(chosen_probs, long_probs)` is returned. If
        `return_long_probs is True` and `chosen_row_to_obs is None`, then
        `long_probs` is returned. If `chosen_row_to_obs` is passed and
        `return_long_probs is False` then `chosen_probs` is returned.

        `chosen_probs` is a 1D numpy array of shape (num_observations,). Each
        element is the probability of the corresponding observation being
        associated with its realized outcome.

        `long_probs` is a 1D numpy array with one element per observation per
        available alternative for that observation. Each element is the
        probability of the corresponding observation being associated with that
        rows corresponding alternative.

        If `beta` is a 2D array, `chosen_probs` and `long_probs` will also be
        2D arrays, with as many columns as there are sets of parameters being
        used to calculate probabilities with.

        It is NOT valid to have `chosen_row_to_obs == None` and
        `return_long_probs == False`.
    """
    # Check argument validity
    if (len(beta.shape) >= 2) and (len(design.shape) >= 3):
        msg_1 = "Cannot calculate probabilities with both 3D design matrix AND"
        msg_2 = " 2D coefficient array."
        raise ValueError(msg_1 + msg_2)
    if chosen_row_to_obs is None and return_long_probs is False:
        msg = "chosen_row_to_obs is None AND return_long_probs is False"
        raise ValueError(msg)

    # Calculate the systematic utility for each alternative for each individual
    sys_utilities = design.dot(beta)

    # Calculate the probability from the transformed utilities
    # The transformed utilities will be of shape (num_rows, 1)
    transformed_utilities = utility_transform(sys_utilities,
                                              alt_IDs,
                                              rows_to_alts,
                                              shape_params,
                                              intercept_params)

    # The following commands are to guard against numeric under/over-flow
    too_small_idx = transformed_utilities < min_exponent_val
    too_large_idx = transformed_utilities > max_exponent_val

    transformed_utilities[too_small_idx] = min_exponent_val
    transformed_utilities[too_large_idx] = max_exponent_val

    # Exponentiate the transformed utilities
    long_exponentials = np.exp(transformed_utilities)

    # long_probs will be of shape (num_rows,) Each element will provide the
    # probability of the observation associated with that row having the
    # alternative associated with that row as the observation's outcome
    individual_denominators = np.asarray(rows_to_obs.transpose().dot(
                                                    long_exponentials))
    long_denominators = np.asarray(rows_to_obs.dot(individual_denominators))
    if len(long_exponentials.shape) > 1 and long_exponentials.shape[1] > 1:
        long_probs = (long_exponentials / long_denominators)
    else:
        long_probs = (long_exponentials / long_denominators).ravel()

    # Guard against underflow
    long_probs[long_probs == 0] = min_comp_value

    if chosen_row_to_obs is None:
        chosen_probs = None
    else:
        # chosen_probs will be of shape (num_observations,)
        chosen_exponentials = np.asarray(
                         chosen_row_to_obs.transpose().dot(long_exponentials))
        if len(long_exponentials.shape) > 1 and long_exponentials.shape[1] > 1:
            chosen_probs = chosen_exponentials / individual_denominators
        else:
            chosen_probs = (chosen_exponentials /
                            individual_denominators).ravel()

    # Return the long form and chosen probabilities if desired
    if return_long_probs and chosen_probs is not None:
        return chosen_probs, long_probs
    # If working with predictions, return just the long form probabilities
    elif return_long_probs and chosen_probs is None:
        return long_probs
    # If estimating the model and storing fitted probabilities or testing the
    # model on data for which we know the chosen alternative, just return the
    # chosen probabilities.
    elif chosen_probs is not None:
        return chosen_probs


def calc_log_likelihood(beta,
                        design,
                        alt_IDs,
                        rows_to_obs,
                        rows_to_alts,
                        choice_vector,
                        utility_transform,
                        intercept_params=None,
                        shape_params=None,
                        PMLE=None,
                        PMLE_lambda=0,
                        weights=None):
    """
    Parameters
    ----------
    beta : 1D ndarray.
        All elements should by ints, floats, or longs. Should have 1 element
        for each utility coefficient being estimated (i.e. num_features).
    design : 2D ndarray.
        There should be one row per observation per available alternative.
        There should be one column per utility coefficient being estimated. All
        elements should be ints, floats, or longs.
    alt_IDs : 1D ndarray.
        All elements should be ints. There should be one row per obervation per
        available alternative for the given observation. Elements denote the
        alternative corresponding to the given row of the design matrix.
    rows_to_obs : 2D ndarray.
        There should be one row per observation per available alternative and
        one column per observation. This matrix maps the rows of the design
        matrix to the unique observations (on the columns).
    rows_to_alts : 2D ndarray.
        There should be one row per observation per available alternative and
        one column per possible alternative. This matrix maps the rows of the
        design matrix to the possible alternatives for this dataset.
    choice_vector : 1D ndarray.
        All elements should be either ones or zeros. There should be one row
        per observation per available alternative for the given observation.
        Elements denote the alternative which is chosen by the given
        observation with a 1 and a zero otherwise.
    utility_transform:  callable.
        Should accept a 1D array of systematic utility values, a 1D array of
        alternative IDs, and miscellaneous args and kwargs. Should return a 1D
        array whose elements contain the appropriately transformed systematic
        utility values, based on the current model being evaluated.
    intercept_params:   1D ndarray, or None, optional.
        If an array, each element should be an int, float, or long. For
        identifiability, there should be J- 1 elements where J is the total
        number of observed alternatives for this dataset. Default == None.
    shape_params: 1D ndarray, or None, optional.
        If an array, each element should be an int, float, or long. There
        should be one value per shape parameter of the model being used.
        Default == None.
    PMLE: None or string value: ['LASSO', 'RIDGE' or 'Tikhonov']
        It determines if a Penalized Maximum Likelihood Estimation should be
        executed. The value of the parameter determines the type of PMLE. The
        None value states that no PMLE is executed. Default = None.
    PMLE_lambda : int, float, long, or None, optional.
        Lambda parameter for LASSO or ridge regression. It should be an int,
        float or long and determines the penalty for the optimization.
        Default = 0.
    weights : 1D ndarray or None, optional.
        Allows for the calculation of weighted log-likelihoods. The weights can
        represent various things. In stratified samples, the weights may be
        the proportion of the observations in a given strata for a sample in
        relation to the proportion of observations in that strata in the
        population. In latent class models, the weights may be the probability
        of being a particular class.

    Returns
    -------
    log_likelihood : float. The log likelihood of the multinomial choice model.
    """
    # Calculate the probability of each individual choosing each available
    # alternative for that individual.
    long_probs = calc_probabilities(beta,
                                    design,
                                    alt_IDs,
                                    rows_to_obs,
                                    rows_to_alts,
                                    utility_transform,
                                    intercept_params=intercept_params,
                                    shape_params=shape_params,
                                    return_long_probs=True)

    # Calculate the weights for the sample
    if weights is None:
        weights = 1

    # Calculate the log likelihood
    log_likelihood = choice_vector.dot(weights * np.log(long_probs))

    if PMLE is None:
        return log_likelihood
    else:
        param_list = [x for x in [shape_params, intercept_params, beta]
                      if x is not None]
        if len(param_list) > 1:
            params = np.concatenate(param_list, axis=0)
        else:
            params = param_list[0]

        if PMLE == "LASSO":
            return log_likelihood - PMLE_lambda * np.absolute(params).sum()

        elif PMLE == "RIDGE":
            return log_likelihood - PMLE_lambda * np.square(params).sum()

        elif PMLE == "Tikhonov":
            # TODO: Not implemented.
            msg = "Tikhonov Penalized Maximum Likelihood Estimation is not yet"
            msg_2 = "implemented. It would be available in a future version."
            msg_3 = "\nUse another PMLE type instead or None for no PMLE."
            total_msg = " ".join([msg_1, msg_2, msg_3])
            raise ValueError(total_msg)

        else:
            msg_1 = "Error! {} is not a valid value for PMLE.\n".format(PMLE)
            msg_2 = "Penalized Maximum Likelihood Estimation (MPLE) only"
            msg_3 = "implements methods: 'LASSO', 'RIDGE' and 'Tikhonov' (only"
            msg_4 = "for kernel models)."
            total_msg = " ".join([msg_1, msg_2, msg_3])
            raise ValueError(total_msg)


def calc_gradient(beta,
                  design,
                  alt_IDs,
                  rows_to_obs,
                  rows_to_alts,
                  choice_vector,
                  utility_transform,
                  transform_first_deriv_c,
                  transform_first_deriv_v,
                  transform_deriv_alpha,
                  intercept_params,
                  shape_params,
                  PMLE,
                  PMLE_lambda,
                  weights):
    """
    Parameters
    ----------
    beta : 1D ndarray.
        All elements should by ints, floats, or longs. Should have 1 element
        for each utility coefficient being estimated (i.e. num_features).
    design : 2D ndarray.
        Tjere should be one row per observation per available alternative.
        There should be one column per utility coefficient being estimated. All
        elements should be ints, floats, or longs.
    alt_IDs : 1D ndarray.
        All elements should be ints. There should be one row per obervation per
        available alternative for the given observation. Elements denote the
        alternative corresponding to the given row of the design matrix.
    rows_to_obs : 2D scipy sparse array.
        There should be one row per observation per available alternative and
        one column per observation. This matrix maps the rows of the design
        matrix to the unique observations (on the columns).
    rows_to_alts : 2D scipy sparse array
        There should be one row per observation per available alternative and
        one column per possible alternative. This matrix maps the rows of the
        design matrix to the possible alternatives for this dataset.
    choice_vector : 1D ndarray.
        All elements should be either ones or zeros. There should be one row
        per observation per available alternative for the given observation.
        Elements denote the alternative which is chosen by the given
        observation with a 1 and a zero otherwise.
    utility_transform : callable.
        Must accept a 1D array of systematic utility values, a 1D array of
        alternative IDs, and miscellaneous args and kwargs. Should return a 1D
        array whose elements contain the appropriately transformed systematic
        utility values, based on the current model being evaluated.
    transform_first_deriv_c : callable.
        Must accept a 1D array of systematic utility values, a 1D array of
        alternative IDs, the `rows_to_alts` array, (shape parameters if there
        are any) and miscellaneous args and kwargs. Should return a 2D matrix
        or sparse array whose elements contain the derivative of the tranformed
        utility vector with respect to the vector of shape parameters. The
        dimensions of the returned vector should be
        `(design.shape[0], num_alternatives)`. If there are no shape parameters
        then the callable should return None.
    transform_first_deriv_v : callable.
        Must accept a 1D array of systematic utility values, a 1D array of
        alternative IDs, (shape parameters if there are any) and miscellaneous
        args and kwargs. Should return a 2D array whose elements contain the
        derivative of the tranformed utility vector with respect to the vector
        of systematic utilities. The dimensions of the returned vector should
        be `(design.shape[0], design.shape[0])`.
    transform_deriv_alpha : callable.
        Must accept a 1D array of systematic utility values, a 1D array of
        alternative IDs, the `rows_to_alts` array, (intercept parameters if
        there are any) and miscellaneous args and kwargs. Should return a 2D
        array whose elements contain the derivative of the tranformed utility
        vector with respect to the vector of shape parameters. The dimensions
        of the returned vector should be
        `(design.shape[0], num_alternatives - 1)`. If there are no intercept
        parameters, the callable should return None.
    intercept_params : 1D numpy array or None.
        If an array, each element should be an int, float, or long. For
        identifiability, there should be J- 1 elements where J is the total
        number of observed alternatives for this dataset. Default == None.
    shape_params : 1D ndarray or None.
       If an array, each element should be an int, float, or long. There should
       be one value per shape parameter of the model being used.
       Default == None.
    PMLE: None or string value: ['LASSO', 'RIDGE' or 'Tikhonov']
        It determines if a Penalized Maximum Likelihood Estimation should be
        executed. The value of the parameter determines the type of PMLE. The
        None value states that no PMLE is executed. Default = None.
    PMLE_lambda : int, float, long, or None, optional.
        Lambda parameter for LASSO or ridge regression. It should be an int,
        float or long and determines the penalty for the optimization.
        Default = 0.
    weights : 1D ndarray or None.
        Allows for the calculation of weighted log-likelihoods. The weights can
        represent various things. In stratified samples, the weights may be
        the proportion of the observations in a given strata for a sample in
        relation to the proportion of observations in that strata in the
        population. In latent class models, the weights may be the probability
        of being a particular class.

    Returns
    -------
    gradient : 1D ndarray.
       It's shape is (beta.shape[0], ). It is the second derivative of the log-
       likelihood with respect to beta.
    """
    # Calculate the systematic utility for each alternative for each individual
    sys_utilities = design.dot(beta)

    # Calculate the probability of each individual choosing each available
    # alternative for that individual.
    long_probs = calc_probabilities(beta,
                                    design,
                                    alt_IDs,
                                    rows_to_obs,
                                    rows_to_alts,
                                    utility_transform,
                                    intercept_params=intercept_params,
                                    shape_params=shape_params,
                                    return_long_probs=True)

    # Calculate the weights for the sample
    if weights is None:
        weights = 1

    ##########
    # Get the required matrices
    ##########
    # Differentiate the transformed utilities with respect to the shape params
    # Note that dh_dc should be a sparse array
    dh_dc = transform_first_deriv_c(sys_utilities, alt_IDs,
                                    rows_to_alts, shape_params)
    # Differentiate the transformed utilities by the intercept params
    # Note that dh_d_alpha should be a sparse array
    dh_d_alpha = transform_deriv_alpha(sys_utilities, alt_IDs,
                                       rows_to_alts, intercept_params)
    # Differentiate the transformed utilities with respect to the systematic
    # utilities. Note that dh_dv should be a sparse matrix
    dh_dv = transform_first_deriv_v(sys_utilities, alt_IDs,
                                    rows_to_alts, shape_params)
    # Differentiate the transformed utilities with respect to the utility
    # coefficients. Note that dh_db should be a dense **matrix**, not a dense
    # 2D array. This is because the dot product of a 2D scipy sparse array and
    # a 2D dense numpy array yields a 2D dense numpy matrix
    dh_db = dh_dv.dot(design)
    # Differentiate the log likelihood w/ respect to the transformed utilities
    # Note that d_ll_dh will be a dense 2D numpy array.
    d_ll_dh = np.multiply(weights, choice_vector - long_probs)[np.newaxis, :]

    # Calculate the gradient of the log-likelihood with respect to the betas
    d_ll_d_beta = d_ll_dh.dot(dh_db)

    ##########
    # Form and return the gradient
    ##########
    if shape_params is not None and intercept_params is not None:
        # Note that we use d_ll_dh * dh_dc and d_ll_dh * dh_d_alpha because
        # that is how one computes the dot product between a dense 2D numpy
        # array and a 2D sparse matrix. This is due to numpy ndarrays and
        # scipy sparse matrices not playing nicely together. However, numpy
        # ndarrays and numpy matrices can be dot producted together,
        # hence d_ll_dh.dot(dh_db).

        # Note that the 'np.asarray' is because dll_dh * dh_dc will be a row
        # matrix, but we want a 1D numpy array.
        gradient = np.concatenate((np.asarray(d_ll_dh * hstack((dh_dc,
                                                                dh_d_alpha),
                                                               format='csr')),
                                   d_ll_d_beta), axis=1).ravel()
        params = np.concatenate((shape_params, intercept_params, beta),
                                axis=0)

    elif shape_params is not None and intercept_params is None:
        # Note that we use d_ll_dh * dh_dc because that is how one computes
        # the dot product between a dense 2D numpy array and a 2D sparse matrix
        # This is due to numpy ndarrays and scipy sparse matrices not playing
        # nicely together. However, numpy ndarrays and numpy matrices can be
        # dot producted together, hence d_ll_dh.dot(dh_db).

        # Note that the 'np.asarray' is because dll_dh * dh_dc will be a row
        # matrix, but we want a 1D numpy array.
        gradient = np.concatenate((np.asarray(d_ll_dh * dh_dc), d_ll_d_beta),
                                  axis=1).ravel()
        params = np.concatenate((shape_params, beta), axis=0)

    elif shape_params is None and intercept_params is not None:
        # Note that we use d_ll_dh * dh_d_alpha because that's how one computes
        # the dot product between a dense 2D numpy array and a 2D sparse matrix
        # This is due to numpy ndarrays and scipy sparse matrices not playing
        # nicely together. However, numpy ndarrays and numpy matrices can be
        # dot producted together, hence d_ll_dh.dot(dh_db).

        # Note 'np.asarray' is used because dll_dh * dh_d_alpha will be a row
        # matrix, but we want a 1D numpy array.
        gradient = np.concatenate((np.asarray(d_ll_dh * dh_d_alpha),
                                   d_ll_d_beta), axis=1).ravel()
        params = np.concatenate((intercept_params, beta), axis=0)

    else:
        gradient = d_ll_d_beta.ravel()
        params = beta

    if PMLE is not None:
        if PMLE == "LASSO":
            gradient -= 2 * PMLE_lambda * np.sign(params)

        elif PMLE == "RIDGE":
            gradient -= 2 * PMLE_lambda * params

        elif PMLE == "Tikhonov":
            # TODO: Not implemented.
            msg = "Tikhonov Penalized Maximum Likelihood Estimation is not yet"
            msg_2 = "implemented. It would be available in a future version."
            msg_3 = "\nUse another PMLE type instead or None for no PMLE."
            total_msg = " ".join([msg_1, msg_2, msg_3])
            raise ValueError(total_msg)

        else:
            msg_1 = "Error! {} is not a valid value for PMLE.\n".format(PMLE)
            msg_2 = "Penalized Maximum Likelihood Estimation (MPLE) only"
            msg_3 = "implements methods: 'LASSO', 'RIDGE' and 'Tikhonov' (only"
            msg_4 = "for kernel models)."
            total_msg = " ".join([msg_1, msg_2, msg_3])
            raise ValueError(total_msg)

    return gradient


##########
# The three functions below are used, jointly, to construct dP_dV, the block
# diagonal matrix that is the derivative of the long probability vector with
# respect to the long vector of index values = X*beta.
##########
def create_matrix_block_indices(row_to_obs):
    """
    Parameters
    ----------
    row_to_obs: 2D ndarray.
        There should be one row per observation per available alternative and
        one column per observation. This matrix maps the rows of the design
        matrix to the unique observations (on the columns).

    Returns
    -------
    output_indices : list of arrays.
        There will be one array per column in `row_to_obs`. The array will note
        which rows correspond to which observations.
    """
    # Initialize the list of index arrays to be returned
    output_indices = []
    # Determine the number of observations in the dataset
    num_obs = row_to_obs.shape[1]
    # Get the indices of the non-zero elements and their values
    row_indices, col_indices, values = scipy.sparse.find(row_to_obs)
    # Iterate over each observation, i.e. each column in row_to_obs, and
    # determine which rows belong to that observation (i.e. the rows with ones
    # in them).
    for col in xrange(num_obs):
        # Store the array of row indices belonging to the current observation
        output_indices.append(row_indices[np.where(col_indices == col)])

    return output_indices


def robust_outer_product(vec_1, vec_2):
    """
    Calculates a 'robust' outer product of two vectors that may or may not
    contain very small values.

    Parameters
    ----------
    vec_1 : 1D ndarray
    vec_2 : 1D ndarray

    Returns
    -------
    outer_prod : 2D ndarray. The outer product of vec_1 and vec_2
    """
    mantissa_1, exponents_1 = np.frexp(vec_1)
    mantissa_2, exponents_2 = np.frexp(vec_2)
    new_mantissas = mantissa_1[None, :] * mantissa_2[:, None]
    new_exponents = exponents_1[None, :] + exponents_2[:, None]
    return new_mantissas * np.exp2(new_exponents)


def create_matrix_blocks(long_probs, matrix_block_indices):
    """
    Parameters
    ----------
    long_probs : 1D ndarray.
        There should be one row per observation per available alternative. The
        elements of the array will indicate the probability of the alternative
        being the outcome associated with the corresponding observation.
    matrix_block_indices : list of arrays.
        There will be one array per observation. The arrays will note which
        rows correspond to which observations.

    Returns
    -------
    output_matrices : list of matrices.
        Each matrix will contain the derivative of P_i with respect to H_i, and
        there will be one matrix for each observations i. P_i is the array of
        probabilities of each observation being associated with its available
        alternatives. H_i is the array of transformed index values for each
        alternative that is available to observation i.
    """
    # Initialize the list of matrices that is to be returned.
    output_matrices = []
    # Iterate over each observation, i.e. over each list of rows that
    # corresponds to each observation.
    for indices in matrix_block_indices:
        # Isolate P_i, the vector of probabilities of each alternative that
        # is associated with the current observation
        current_probs = long_probs[indices]
        # Get the outer product of the current probabilities
        # probability_outer_product = np.outer(current_probs, current_probs)
        probability_outer_product = robust_outer_product(current_probs,
                                                         current_probs)

        # Create the desired dP_i/dh_i matrix
        dP_i_dh_i = np.diag(current_probs) - probability_outer_product
        # Ensure that the diagonal is positive and non-zero, since it must be.
        diag_idx = np.diag_indices_from(dP_i_dh_i)
        diag_values = dP_i_dh_i[diag_idx].copy()
        diag_values[diag_values == 0] = min_comp_value
        dP_i_dh_i[diag_idx] = diag_values
        # Guard against underflow on the off-diagonals
        underflow_idxs = np.where(dP_i_dh_i == 0)
        for i in xrange(underflow_idxs[0].size):
            row_idx, col_idx = underflow_idxs[0][i], underflow_idxs[1][i]
            if row_idx != col_idx:
                # Since this type of underflow essentially comes from
                # multiplying two very small numbers, just set the overall
                # result to a small number
                dP_i_dh_i[row_idx,
                          col_idx] = -1 * min_comp_value
        # Store the desired dP_i/dh_i matrix
        output_matrices.append(dP_i_dh_i)

    return output_matrices


def calc_hessian(beta,
                 design,
                 alt_IDs,
                 rows_to_obs,
                 rows_to_alts,
                 utility_transform,
                 transform_first_deriv_c,
                 transform_first_deriv_v,
                 transform_deriv_alpha,
                 block_matrix_idxs,
                 intercept_params,
                 shape_params,
                 PMLE,
                 PMLE_lambda,
                 weights):
    """
    Parameters
    ----------
    beta : 1D ndarray.
        All elements should by ints, floats, or longs. Should have 1 element
        for each utility coefficient being estimated (i.e. num_features).
    design : 2D ndarray.
        There should be one row per observation per available alternative.
        There should be one column per utility coefficient being estimated. All
        elements should be ints, floats, or longs.
    alt_IDs : 1D ndarray.
        All elements should be ints. There should be one row per obervation per
        available alternative for the given observation. Elements denote the
        alternative corresponding to the given row of the design matrix.
    rows_to_obs : 2D ndarray.
        There should be one row per observation per available alternative and
        one column per observation. This matrix maps the rows of the design
        matrix to the unique observations (on the columns).
    rows_to_alts: 2D ndarray.
        There should be one row per observation per available alternative and
        one column per possible alternative. This matrix maps the rows of the
        design matrix to the possible alternatives for this dataset.
    utility_transform : callable.
        Must accept a 1D array of systematic utility values, a 1D array of
        alternative IDs, and miscellaneous args and kwargs. Should return a 1D
        array whose elements contain the appropriately transformed systematic
        utility values, based on the current model being evaluated.
    transform_first_deriv_c : callable.
        Must accept a 1D array of systematic utility values, a 1D array of
        alternative IDs, the `rows_to_alts` array, (shape parameters if there
        are any) and miscellaneous args and kwargs. Should return a 2D array
        whose elements contain the derivative of the tranformed utilities with
        respect to the vector of shape parameters. The dimensions of the
        returned vector should be `(design.shape[0], num_alternatives)`.
    transform_first_deriv_v : callable.
        Must accept a 1D array of systematic utility values, a 1D array of
        alternative IDs, (shape parameters if there are any) and miscellaneous
        args and kwargs. Should return a 2D array whose elements contain the
        derivative of the tranformed utility vector with respect to the vector
        of systematic utilities. The dimensions of the returned vector should
        be `(design.shape[0], design.shape[0])`.
    transform_deriv_alpha : callable.
        Must accept a 1D array of systematic utility values, a 1D array of
        alternative IDs, the rows_to_alts array, (intercept parameters if
        there are any) and miscellaneous args and kwargs. Should return a 2D
        array whose elements contain the derivative of the tranformed utilities
        with respect to the vector of shape parameters. The dimensions of the
        returned vector should be `(design.shape[0], num_alternatives - 1)`. If
        `intercept_params == None`, the callable should return None.
    block_matrix_idxs : list of arrays.
        There will be one array per observation. The arrays will note which
        rows correspond to which observations.
    intercept_params : 1D ndarray.
        Each element should be an int, float, or long. For identifiability,
        there should be J- 1 elements where J is the total number of observed
        alternatives in the dataset.
    shape_params: None or 1D ndarray.
        If an array, each element should be an int, float, or long. There
        should be one value per shape parameter of the model being used.
        Default == None.
    PMLE: None or string value: ['LASSO', 'RIDGE' or 'Tikhonov']
        It determines if a Penalized Maximum Likelihood Estimation should be
        executed. The value of the parameter determines the type of PMLE. The
        None value states that no PMLE is executed. Default = None.
    PMLE_lambda : int, float, long, or None, optional.
        Lambda parameter for LASSO or ridge regression. It should be an int,
        float or long and determines the penalty for the optimization.
        Default = 0.
    weights : 1D ndarray or None.
        Allows for the calculation of weighted log-likelihoods. The weights can
        represent various things. In stratified samples, the weights may be
        the proportion of the observations in a given strata for a sample in
        relation to the proportion of observations in that strata in the
        population. In latent class models, the weights may be the probability
        of being a particular class.

    Returns
    -------
    hess : 2D ndarray.
        It's shape is `(beta.shape[0], beta.shape[0])`. It is the second
        derivative of the log likelihood with respect to beta.
    """
    # Calculate the systematic utility for each alternative for each individual
    sys_utilities = design.dot(beta)

    # Calculate the probability of each individual choosing each available
    # alternative for that individual.
    long_probs = calc_probabilities(beta,
                                    design,
                                    alt_IDs,
                                    rows_to_obs,
                                    rows_to_alts,
                                    utility_transform,
                                    intercept_params=intercept_params,
                                    shape_params=shape_params,
                                    return_long_probs=True)

    # Calculate the weights for the sample
    if weights is None:
        weights = np.ones(design.shape[0])

    ##########
    # Get the required matrices
    ##########
    # Differentiate the transformed utilities with respect to the shape params
    # Note that dh_dc will be a 2D scipy sparse matrix
    dh_dc = transform_first_deriv_c(sys_utilities, alt_IDs,
                                    rows_to_alts, shape_params)
    # Differentiate the transformed utilities with respect to the systematic
    # utilities. Note that dh_dv will be a 2D scipy sparse matrix.
    dh_dv = transform_first_deriv_v(sys_utilities, alt_IDs,
                                    rows_to_alts, shape_params)
    # Differentiate the transformed utilities by the intercept params
    # Note that dh_d_alpha should be a sparse array
    dh_d_alpha = transform_deriv_alpha(sys_utilities, alt_IDs,
                                       rows_to_alts, intercept_params)
    # Differentiate the transformed utilities with respect to the utility
    # coefficients. Note that dh_db will be a 2D dense numpy matrix
    dh_db = dh_dv.dot(design)

    # Differentiate the probabilities with respect to the transformed utilities
    # Note that dp_dh will be a 2D dense numpy array
    block_matrices = create_matrix_blocks(long_probs, block_matrix_idxs)
    dp_dh = block_diag(*block_matrices) * weights[None, :]

    ##########
    # Calculate the second and mixed partial derivatives within the hessian
    ##########
    # Calculate the second derivative of the log-likelihood with repect to the
    # utility coefficients. Should have shape (design.shape[0],
    # design.shape[0]). Since dp_dh is a 2D dense numpy array and dh_db is a
    # 2D dense numpy matrix, using the .dot() syntax should work to compute
    # the dot product.
    d2_ll_db2 = -1 * dh_db.T.dot(dp_dh.dot(dh_db))

    ##########
    # Form and return the hessian
    ##########
    if shape_params is not None and intercept_params is not None:
        # Calculate the second derivative of the log-likelihood with respect
        # to the shape parameters. Should have shape (shape_params.shape[0],
        # shape_params.shape[0]). Note that since dp_dh is a 2D dense numpy
        # array and dh_dc is a sparse matrix or dense numpy matrix, to
        # compute the dot product we need to use the * operator
        d2_ll_dc2 = -1 * dh_dc.T.dot(dp_dh * dh_dc)

        # Calculate the second derivative of the log-likelihood with respect
        # to the intercept parameters. Should have shape (J - 1, J - 1) where
        # J is the total number of observed alternatives for this dataset.
        # Note that since dp_dh is a 2D dense numpy array and dh_d_alpha is a
        # sparse matrix or dense numpy matrix, to compute the dot product
        # we need to use the * operator
        d2_ll_d_alpha2 = -1 * dh_d_alpha.T.dot(dp_dh * dh_d_alpha)

        # Calculate the mixed second derivative of the log-likelihood with
        # respect to the intercept and shape parameters. Should have shape
        # (dh_d_alpha.shape[1], dh_dc.shape[1]). Note that since dp_dh is a 2D
        # dense numpy array and dh_dc is a sparse matrix or dense numpy
        # matrix, to compute the dot product we need to use the * operator
        d2_ll_dc_d_alpha = -1 * dh_d_alpha.T.dot(dp_dh * dh_dc)

        # Calculate the mixed partial derivative of the log-likelihood with
        # respect to the utility coefficients and then with respect to the
        # shape parameters. Should have shape (design.shape[0],
        # shape_params.shape[0]). Note that since dp_dh is a 2D dense numpy
        # array and dh_dc is a sparse matrix or dense numpy matrix, to
        # compute the dot product we need to use the * operator
        d2_ll_dc_db = -1 * dh_db.T.dot(dp_dh * dh_dc)

        # Calculate the mixed partial derivative of the log-likelihood with
        # respect to the utility coefficients and then with respect to the
        # intercept parameters. Should have shape (design.shape[0],
        # intercept_params.shape[0]). Note that since dp_dh is a 2D dense numpy
        # array and dh_d_alpha is a sparse matrix or dense numpy matrix, to
        # compute the dot product we need to use the * operator
        d2_ll_d_alpha_db = -1 * dh_db.T.dot(dp_dh * dh_d_alpha)

        # Form the 3 by 3 partitioned hessian of 2nd derivatives
        top_row = np.concatenate((d2_ll_dc2,
                                  d2_ll_dc_d_alpha.T,
                                  d2_ll_dc_db.T), axis=1)
        middle_row = np.concatenate((d2_ll_dc_d_alpha,
                                     d2_ll_d_alpha2,
                                     d2_ll_d_alpha_db.T), axis=1)
        last_row = np.concatenate((d2_ll_dc_db,
                                   d2_ll_d_alpha_db,
                                   d2_ll_db2), axis=1)
        hess = np.concatenate((top_row,
                               middle_row,
                               last_row), axis=0)

    elif shape_params is not None and intercept_params is None:
        # Calculate the second derivative of the log-likelihood with respect
        # to the shape parameters. Should have shape (shape_params.shape[0],
        # shape_params.shape[0]). Note that since dp_dh is a 2D dense numpy
        # array and dh_dc is a sparse matrix or dense numpy matrix, to
        # compute the dot product we need to use the * operator
        d2_ll_dc2 = -1 * dh_dc.T.dot(dp_dh * dh_dc)

        # Calculate the mixed partial derivative of the log-likelihood with
        # respect to the utility coefficients and then with respect to the
        # shape parameters. Should have shape (design.shape[0],
        # shape_params.shape[0]). Note that since dp_dh is a 2D dense numpy
        # array and dh_dc is a sparse matrix or dense numpy matrix, to
        # compute the dot product we need to use the * operator
        d2_ll_dc_db = -1 * dh_db.T.dot(dp_dh * dh_dc)

        hess = np.concatenate((np.concatenate((d2_ll_dc2,
                                               d2_ll_dc_db.T), axis=1),
                               np.concatenate((d2_ll_dc_db,
                                               d2_ll_db2), axis=1)), axis=0)

    elif shape_params is None and intercept_params is not None:
        # Calculate the second derivative of the log-likelihood with respect
        # to the intercept parameters. Should have shape (J - 1, J - 1) where
        # J is the total number of observed alternatives for this dataset.
        # Note that since dp_dh is a 2D dense numpy array and dh_d_alpha is a
        # sparse matrix or dense numpy matrix, to compute the dot product
        # we need to use the * operator
        d2_ll_d_alpha2 = -1 * dh_d_alpha.T.dot(dp_dh * dh_d_alpha)

        # Calculate the mixed partial derivative of the log-likelihood with
        # respect to the utility coefficients and then with respect to the
        # intercept parameters. Should have shape (design.shape[0],
        # intercept_params.shape[0]). Note that since dp_dh is a 2D dense numpy
        # array and dh_d_alpha is a sparse matrix or dense numpy matrix, to
        # compute the dot product we need to use the * operator
        d2_ll_d_alpha_db = -1 * dh_db.T.dot(dp_dh * dh_d_alpha)

        hess = np.concatenate((np.concatenate((d2_ll_d_alpha2,
                                               d2_ll_d_alpha_db.T), axis=1),
                               np.concatenate((d2_ll_d_alpha_db,
                                               d2_ll_db2), axis=1)), axis=0)
    else:
        hess = d2_ll_db2

    if PMLE is not None:
        if PMLE == "LASSO":
            # Do nothing because the expression should be hess -= 0
            pass

        elif PMLE == "RIDGE":
            hess -= 2 * PMLE_lambda

        elif PMLE == "Tikhonov":
            # TODO: Not implemented.
            msg = "Tikhonov Penalized Maximum Likelihood Estimation is not yet"
            msg_2 = "implemented. It would be available in a future version."
            msg_3 = "\nUse another PMLE type instead or None for no PMLE."
            total_msg = " ".join([msg_1, msg_2, msg_3])
            raise ValueError(total_msg)

        else:
            msg_1 = "Error! {} is not a valid value for PMLE.\n".format(PMLE)
            msg_2 = "Penalized Maximum Likelihood Estimation (MPLE) only"
            msg_3 = "implements methods: 'LASSO', 'RIDGE' and 'Tikhonov' (only"
            msg_4 = "for kernel models)."
            total_msg = " ".join([msg_1, msg_2, msg_3])
            raise ValueError(total_msg)

    # Make sure we are returning standard numpy arrays
    if isinstance(hess, np.matrixlib.defmatrix.matrix):
        hess = np.asarray(hess)

    return hess


def calc_fisher_info_matrix(beta,
                            design,
                            alt_IDs,
                            rows_to_obs,
                            rows_to_alts,
                            choice_vector,
                            utility_transform,
                            transform_first_deriv_c,
                            transform_first_deriv_v,
                            transform_deriv_alpha,
                            intercept_params,
                            shape_params,
                            PMLE,
                            PMLE_lambda,
                            weights):
    """
    Parameters
    ----------
    beta : 1D ndarray.
        All elements should by ints, floats, or longs. Should have 1 element
        for each utility coefficient being estimated (i.e. num_features).
    design : 2D ndarray.
        There should be one row per observation per available alternative.
        There should be one column per utility coefficient being estimated. All
        elements should be ints, floats, or longs.
    alt_IDs : 1D ndarray.
        All elements should be ints. There should be one row per obervation per
        available alternative for the given observation. Elements denote the
        alternative corresponding to the given row of the design matrix.
    rows_to_obs : 2D ndarray.
        There should be one row per observation per available alternative and
        one column per observation. This matrix maps the rows of the design
        matrix to the unique observations (on the columns).
    rows_to_alts : 2D ndarray
        There should be one row per observation per available alternative and
        one column per possible alternative. This matrix maps the rows of the
        design matrix to the possible alternatives for this dataset.
    choice_vector : 1D ndarray.
        All elements should be either ones or zeros. There should be one row
        per observation per available alternative for the given observation.
        Elements denote the alternative which is chosen by the given
        observation with a 1 and a zero otherwise.
    utility_transform : callable.
        Must accept a 1D array of systematic utility values, a 1D array of
        alternative IDs, and miscellaneous args and kwargs. Should return a
        1D array whose elements contain the appropriately transformed
        systematic utility values, based on the current model being evaluated.
    transform_first_deriv_c : callable.
        Must accept a 1D array of systematic utility values, a 1D array of
        alternative IDs, the `rows_to_alts` array, (shape parameters if there
        are any) and miscellaneous args and kwargs. Should return a 2D array
        whose elements contain the derivative of the tranformed utilities with
        respect to the vector of shape parameters. The dimensions of the
        returned vector should be `(design.shape[0], num_alternatives)`.
    transform_first_deriv_v : callable.
        Must accept a 1D array of systematic utility values, a 1D array of
         alternative IDs, (shape parameters if there are any) and miscellaneous
         args and kwargs. Should return a 2D array whose elements contain the
         derivative of the utility tranformation vector with respect to the
         vector of systematic utilities. The dimensions of the returned vector
         should be `(design.shape[0],  design.shape[0])`.
    shape_params : None or 1D ndarray.
        If an array, each element should be an int, float, or long. There
        should be one value per shape parameter of the model being used.
        Default == None.
    PMLE: None or string value: ['LASSO', 'RIDGE' or 'Tikhonov']
        It determines if a Penalized Maximum Likelihood Estimation should be
        executed. The value of the parameter determines the type of PMLE. The
        None value states that no PMLE is executed. Default = None.
    PMLE_lambda : int, float, long, or None, optional.
        Lambda parameter for LASSO or ridge regression. It should be an int,
        float or long and determines the penalty for the optimization.
        Default = 0.
    weights : 1D ndarray or None.
        Allows for the calculation of weighted log-likelihoods. The weights can
        represent various things. In stratified samples, the weights may be
        the proportion of the observations in a given strata for a sample in
        relation to the proportion of observations in that strata in the
        population. In latent class models, the weights may be the probability
        of being a particular class.

    Returns
    -------
    fisher_matrix : 2D numpy array.
        It will be a square matrix, with one row and one column for each shape,
        intercept, and index coefficient. Contains the BHHH approximation to
        the Fisher Information matrix of the log likelihood.
    """
    # Calculate the systematic utility for each alternative for each individual
    sys_utilities = design.dot(beta)

    # Calculate the probability that the individual associated with a given row
    # chooses the alternative associated with a given row.
    long_probs = calc_probabilities(beta,
                                    design,
                                    alt_IDs,
                                    rows_to_obs,
                                    rows_to_alts,
                                    utility_transform,
                                    intercept_params=intercept_params,
                                    shape_params=shape_params,
                                    return_long_probs=True)

    # Calculate the weights for the sample
    if weights is None:
        weights = np.ones(design.shape[0])
    weights_per_obs =\
        np.max(rows_to_obs.toarray() * weights[:, None], axis=0)

    ##########
    # Get the required matrices
    ##########
    # Differentiate the transformed utilities with respect to the shape params
    dh_dc = transform_first_deriv_c(sys_utilities, alt_IDs,
                                    rows_to_alts, shape_params)
    # Differentiate the transformed utilities with respect to the systematic
    # utilities
    dh_dv = transform_first_deriv_v(sys_utilities, alt_IDs,
                                    rows_to_alts, shape_params)
    # Differentiate the transformed utilities by the intercept params
    # Note that dh_d_alpha should be a sparse array
    dh_d_alpha = transform_deriv_alpha(sys_utilities, alt_IDs,
                                       rows_to_alts, intercept_params)
    # Differentiate the transformed utilities with respect to the utility
    # coefficients. This should be a dense numpy array.
    dh_db = np.asarray(dh_dv.dot(design))
    # Differentiate the log likelihood w/ respect to the transformed utilities
    d_ll_dh = (choice_vector - long_probs)[np.newaxis, :]

    ##########
    # Create the matrix where each row represents the gradient of a particular
    # observations log-likelihood with respect to the shape parameters and
    # beta, depending on whether there are shape parameters being estimated
    ##########
    if shape_params is not None and intercept_params is not None:
        if isinstance(dh_dc, np.matrixlib.defmatrix.matrix):
            # Note that the '.A' transforms the matrix into a numpy ndarray
            gradient_vec = d_ll_dh.T * np.concatenate((dh_dc.A,
                                                       dh_d_alpha.toarray(),
                                                       dh_db), axis=1)
        else:
            gradient_vec = d_ll_dh.T * np.concatenate((dh_dc.toarray(),
                                                       dh_d_alpha.toarray(),
                                                       dh_db), axis=1)
    elif shape_params is not None and intercept_params is None:
        if isinstance(dh_dc, np.matrixlib.defmatrix.matrix):
            # Note that the '.A' transforms the matrix into a numpy ndarray
            gradient_vec = d_ll_dh.T * np.concatenate((dh_dc.A, dh_db), axis=1)
        else:
            gradient_vec = d_ll_dh.T * np.concatenate((dh_dc.toarray(),
                                                       dh_db), axis=1)
    elif shape_params is None and intercept_params is not None:
        # Note '.to_array()' is used because dh_d_alpha will be a sparse
        # matrix, but we want a 2D numpy array.
        gradient_vec = d_ll_dh.T * np.concatenate((dh_d_alpha.toarray(),
                                                   dh_db), axis=1)
    else:
        gradient_vec = d_ll_dh.T * dh_db

    # Make sure that we calculate the gradient PER OBSERVATION
    # and then take the outer products of those gradients.
    # Note that this is different than taking the outer products of the
    # gradient of the log-likelihood per available alternative per observation
    gradient_vec = rows_to_obs.T.dot(gradient_vec)

    # Compute and return the outer product of each row of the gradient
    # with itself. Then sum these individual matrices together. The line below
    # does the same computation just with less memory and time.
    fisher_matrix =\
        gradient_vec.T.dot(np.multiply(weights_per_obs[:, None], gradient_vec))

    if PMLE is not None:
        if PMLE == "LASSO":
            # Do nothing because the expression should be fisher_matrix -= 0
            # The rational behind adding 0 is that the fisher information
            # matrix should approximate the hessian and in the hessian we add
            # 0 at the end. I don't know if this is the correct way to
            # calculate the Fisher Information in ridge regression models.
            pass

        elif PMLE == "RIDGE":
            # The rational behind adding 2*ridge is that the fisher information
            # matrix should approximate the hessian and in the hessian we add
            # 2*ridge at the end. I don't know if this is the correct way to
            # calculate the Fisher Information in ridge regression models.
            fisher_matrix -= 2 * PMLE_lambda

        elif PMLE == "Tikhonov":
            # TODO: Not implemented.
            msg = "Tikhonov Penalized Maximum Likelihood Estimation is not yet"
            msg_2 = "implemented. It would be available in a future version."
            msg_3 = "\nUse another PMLE type instead or None for no PMLE."
            total_msg = " ".join([msg_1, msg_2, msg_3])
            raise ValueError(total_msg)

        else:
            msg_1 = "Error! {} is not a valid value for PMLE.\n".format(PMLE)
            msg_2 = "Penalized Maximum Likelihood Estimation (MPLE) only"
            msg_3 = "implements methods: 'LASSO', 'RIDGE' and 'Tikhonov' (only"
            msg_4 = "for kernel models)."
            total_msg = " ".join([msg_1, msg_2, msg_3, msg_4])
            raise ValueError(total_msg)

    return fisher_matrix


def calc_asymptotic_covariance(hessian, fisher_info_matrix):
    """
    Parameters
    ----------
    hessian : 2D ndarray.
        It should have shape `(num_vars, num_vars)`. It is the matrix of second
        derivatives of the total loss across the dataset, with respect to each
        pair of coefficients being estimated.
    fisher_info_matrix : 2D ndarray.
        It should have a shape of `(num_vars, num_vars)`.  It is the
        approximation of the negative of the expected hessian formed by taking
        the outer product of (each observation's gradient of the loss function)
        with itself, and then summing across all observations.

    Returns
    -------
    huber_white_matrix : 2D ndarray.
        Will have shape `(num_vars, num_vars)`. The entries in the returned
        matrix are calculated by the following formula:
        `hess_inverse * fisher_info_matrix * hess_inverse`.
    """
    # Calculate the inverse of the hessian
    hess_inv = scipy.linalg.inv(hessian)

    return np.dot(hess_inv, np.dot(fisher_info_matrix, hess_inv))



"""
@author:    Timothy Brathwaite
@name:      Bootstrap Sampler
@summary:   This module provides functions that will perform the stratified
            resampling needed for the bootstrapping procedure.
"""
from collections import OrderedDict

import numpy as np
import pandas as pd


def relate_obs_ids_to_chosen_alts(obs_id_array,
                                  alt_id_array,
                                  choice_array):
    """
    Creates a dictionary that relates each unique alternative id to the set of
    observations ids that chose the given alternative.

    Parameters
    ----------
    obs_id_array : 1D ndarray of ints.
        Should be a long-format array of observation ids. Each element should
        correspond to the unique id of the unit of observation that corresponds
        to the given row of the long-format data. Note that each unit of
        observation may have more than one associated choice situation.
    alt_id_array : 1D ndarray of ints.
        Should be a long-format array of alternative ids. Each element should
        denote the unique id of the alternative that corresponds to the given
        row of the long format data.
    choice_array : 1D ndarray of ints.
        Each element should be either a one or a zero, indicating whether the
        alternative on the given row of the long format data was chosen or not.

    Returns
    -------
    chosen_alts_to_obs_ids : dict.
        Each key will be a unique value from `alt_id_array`. Each key's value
        will be a 1D ndarray that contains the sorted, unique observation ids
        of those observational units that chose the given alternative.
    """
    # Figure out which units of observation chose each alternative.
    chosen_alts_to_obs_ids = {}

    for alt_id in np.sort(np.unique(alt_id_array)):
        # Determine which observations chose the current alternative.
        selection_condition =\
            np.where((alt_id_array == alt_id) & (choice_array == 1))

        # Store the sorted, unique ids that chose the current alternative.
        chosen_alts_to_obs_ids[alt_id] =\
            np.sort(np.unique(obs_id_array[selection_condition]))

    # Return the desired dictionary.
    return chosen_alts_to_obs_ids


def get_num_obs_choosing_each_alternative(obs_per_alt_dict):
    """
    Will create an ordered dictionary that records the number of units of
    observation that have chosen the given alternative (i.e. the associated
    dictionary key). Will also determine the total number of unique
    observations in the dataset.

    Parameters
    ----------
    obs_per_alt_dict : dict.
        Each key should be a unique alternave id. Each key's value will be 1D
        ndarray that contains the sorted, unique observation ids of those
        observational units that chose the given alternative.

    Returns
    -------
    num_obs_per_group : OrderedDict.
        Keys will be the alternative ids present in `obs_per_alt_dict`. Values
        will be the `len(obs_per_alt_dict[alt_id]).`
    tot_num_obs : int.
        Denotes the total number of unique observation ids in one's dataset.
    """
    # Initialize the object that is to be returned.
    num_obs_per_group = OrderedDict()

    # Determine the number of unique units of observation per group.
    for alt_id in obs_per_alt_dict:
        num_obs_per_group[alt_id] = len(obs_per_alt_dict[alt_id])

    # Determine the total number of units of observation that will be chosen
    # for each bootstrap sample.
    tot_num_obs = sum([num_obs_per_group[g] for g in num_obs_per_group])

    # Return the desired objects.
    return num_obs_per_group, tot_num_obs


def create_cross_sectional_bootstrap_samples(obs_id_array,
                                             alt_id_array,
                                             choice_array,
                                             num_samples,
                                             seed=None):
    """
    Determines the unique observations that will be present in each bootstrap
    sample. This function DOES NOT create the new design matrices or a new
    long-format dataframe for each bootstrap sample. Note that these will be
    correct bootstrap samples for cross-sectional datasets. This function will
    not work correctly for panel datasets.

    Parameters
    ----------
    obs_id_array : 1D ndarray of ints.
        Each element should denote a unique observation id for the
        corresponding row of the long format array.
    alt_id_array : 1D ndarray of ints.
        Each element should denote a unique alternative id for the
        corresponding row of the long format array.
    choice_array : 1D ndarray of ints.
        Each element should be a one or a zero. The values should denote a
        whether or not the corresponding alternative in `alt_id_array` was
        chosen by the observational unit in the corresponding row of
        `obs_id_array.`
    num_samples : int.
        Denotes the number of bootstrap samples that need to be drawn.
    seed : non-negative int or None, optional.
        Denotes the random seed to be used in order to ensure reproducibility
        of the bootstrap sample generation. Default is None. If None, no seed
        will be used and the generation of the bootstrap samples will (in
        general) not be reproducible.


    Returns
    -------
    ids_per_sample : 2D ndarray.
        Each row represents a complete bootstrap sample. Each column denotes a
        selected bootstrap observation that comprises the bootstrap sample. The
        elements of the array denote the observation ids of the chosen
        observational units.
    """
    # Determine the units of observation that chose each alternative.
    chosen_alts_to_obs_ids =\
        relate_obs_ids_to_chosen_alts(obs_id_array, alt_id_array, choice_array)

    # Determine the number of unique units of observation per group and overall
    num_obs_per_group, tot_num_obs =\
        get_num_obs_choosing_each_alternative(chosen_alts_to_obs_ids)

    # Initialize the array that will store the observation ids for each sample
    ids_per_sample = np.empty((num_samples, tot_num_obs), dtype=float)

    if seed is not None:
        # Check the validity of the seed argument.
        if not isinstance(seed, int):
            msg = "`boot_seed` MUST be an int."
            raise ValueError(msg)

        # If desiring reproducibility, set the random seed within numpy
        np.random.seed(seed)

    # Initialize a variable to keep track of what column we're on.
    col_idx = 0
    for alt_id in num_obs_per_group:
        # Get the set of observations that chose the current alternative.
        relevant_ids = chosen_alts_to_obs_ids[alt_id]
        # Determine the number of needed resampled ids.
        resample_size = num_obs_per_group[alt_id]
        # Resample, with replacement, observations who chose this alternative.
        current_ids = (np.random.choice(relevant_ids,
                                        size=resample_size * num_samples,
                                        replace=True)
                                .reshape((num_samples, resample_size)))
        # Determine the last column index to use when storing the resampled ids
        end_col = col_idx + resample_size
        # Assign the sampled ids to the correct columns of ids_per_sample
        ids_per_sample[:, col_idx:end_col] = current_ids
        # Update the column index
        col_idx += resample_size

    # Return the resampled observation ids.
    return ids_per_sample


def create_bootstrap_id_array(obs_id_per_sample):
    """
    Creates a 2D ndarray that contains the 'bootstrap ids' for each replication
    of each unit of observation that is an the set of bootstrap samples.

    Parameters
    ----------
    obs_id_per_sample : 2D ndarray of ints.
        Should have one row for each bootsrap sample. Should have one column
        for each observational unit that is serving as a new bootstrap
        observational unit.

    Returns
    -------
    bootstrap_id_array : 2D ndarray of ints.
        Will have the same shape as `obs_id_per_sample`. Each element will
        denote the fake observational id in the new bootstrap dataset.
    """
    # Determine the shape of the object to be returned.
    n_rows, n_cols = obs_id_per_sample.shape
    # Create the array of bootstrap ids.
    bootstrap_id_array =\
        np.tile(np.arange(n_cols) + 1, n_rows).reshape((n_rows, n_cols))
    # Return the desired object
    return bootstrap_id_array


def create_deepcopied_groupby_dict(orig_df, obs_id_col):
    """
    Will create a dictionary where each key corresponds to a unique value in
    `orig_df[obs_id_col]` and each value corresponds to all of the rows of
    `orig_df` where `orig_df[obs_id_col] == key`.

    Parameters
    ----------
    orig_df : pandas DataFrame.
        Should be long-format dataframe containing the data used to estimate
        the desired choice model.
    obs_id_col : str.
        Should be a column name within `orig_df`. Should denote the original
        observation id column.

    Returns
    -------
    groupby_dict : dict.
        Each key will be a unique value in `orig_df[obs_id_col]` and each value
        will be the rows of `orig_df` where `orig_df[obs_id_col] == key`.
    """
    # Get the observation id values
    obs_id_vals = orig_df[obs_id_col].values
    # Get the unique observation ids
    unique_obs_ids = np.unique(obs_id_vals)
    # Initialize the dictionary to be returned.
    groupby_dict = {}
    # Populate the dictionary with dataframes for each individual.
    for obs_id in unique_obs_ids:
        # Filter out only the rows corresponding to the current observation id.
        desired_rows = obs_id_vals == obs_id
        # Add the desired dataframe to the dictionary.
        groupby_dict[obs_id] = orig_df.loc[desired_rows].copy(deep=True)

    # Return the desired object.
    return groupby_dict


def check_column_existence(col_name, df, presence=True):
    """
    Checks whether or not `col_name` is in `df` and raises a helpful error msg
    if the desired condition is not met.

    Parameters
    ----------
    col_name : str.
        Should represent a column whose presence in `df` is to be checked.
    df : pandas DataFrame.
        The dataframe that will be checked for the presence of `col_name`.
    presence : bool, optional.
        If True, then this function checks for the PRESENCE of `col_name` from
        `df`. If False, then this function checks for the ABSENCE of
        `col_name` in `df`. Default == True.

    Returns
    -------
    None.
    """
    if presence:
        if col_name not in df.columns:
            msg = "Ensure that `{}` is in `df.columns`."
            raise ValueError(msg.format(col_name))
    else:
        if col_name in df.columns:
            msg = "Ensure that `{}` is not in `df.columns`."
            raise ValueError(msg.format(col_name))
    return None


def ensure_resampled_obs_ids_in_df(resampled_obs_ids, orig_obs_id_array):
    """
    Checks whether all ids in `resampled_obs_ids` are in `orig_obs_id_array`.
    Raises a helpful ValueError if not.

    Parameters
    ----------
    resampled_obs_ids : 1D ndarray of ints.
        Should contain the observation ids of the observational units that will
        be used in the current bootstrap sample.
    orig_obs_id_array : 1D ndarray of ints.
        Should countain the observation ids of the observational units in the
        original dataframe containing the data for this model.

    Returns
    -------
    None.
    """
    if not np.in1d(resampled_obs_ids, orig_obs_id_array).all():
        msg =\
            "All values in `resampled_obs_ids` MUST be in `orig_obs_id_array`."
        raise ValueError(msg)
    return None


def create_bootstrap_dataframe(orig_df,
                               obs_id_col,
                               resampled_obs_ids_1d,
                               groupby_dict,
                               boot_id_col="bootstrap_id"):
    """
    Will create the altered dataframe of data needed to estimate a choice model
    with the particular observations that belong to the current bootstrap
    sample.

    Parameters
    ----------
    orig_df : pandas DataFrame.
        Should be long-format dataframe containing the data used to estimate
        the desired choice model.
    obs_id_col : str.
        Should be a column name within `orig_df`. Should denote the original
        observation id column.
    resampled_obs_ids_1d : 1D ndarray of ints.
        Each value should represent the alternative id of a given bootstrap
        replicate.
    groupby_dict : dict.
        Each key will be a unique value in `orig_df[obs_id_col]` and each value
        will be the rows of `orig_df` where `orig_df[obs_id_col] == key`.
    boot_id_col : str, optional.
        Denotes the new column that will be created to specify the bootstrap
        observation ids for choice model estimation.

    Returns
    -------
    bootstrap_df : pandas Dataframe.
        Will contain all the same columns as `orig_df` as well as the
        additional `boot_id_col`. For each value in `resampled_obs_ids_1d`,
        `bootstrap_df` will contain the long format rows from `orig_df` that
        have the given observation id.
    """
    # Check the validity of the passed arguments.
    check_column_existence(obs_id_col, orig_df, presence=True)
    check_column_existence(boot_id_col, orig_df, presence=False)
    # Alias the observation id column
    obs_id_values = orig_df[obs_id_col].values
    # Check the validity of the resampled observation ids.
    ensure_resampled_obs_ids_in_df(resampled_obs_ids_1d, obs_id_values)

    # Initialize a list to store the component dataframes that will be
    # concatenated to form the final bootstrap_df
    component_dfs = []

    # Populate component_dfs
    for boot_id, obs_id in enumerate(resampled_obs_ids_1d):
        # Extract the dataframe that we desire.
        extracted_df = groupby_dict[obs_id].copy()
        # Add the bootstrap id value.
        extracted_df[boot_id_col] = boot_id + 1
        # Store the component dataframe
        component_dfs.append(extracted_df)

    # Create and return the desired dataframe.
    bootstrap_df = pd.concat(component_dfs, axis=0, ignore_index=True)
    return bootstrap_df



# -*- coding: utf-8 -*-
"""
@author:    Timothy Brathwaite
            José Ángel Martín Baos
@summary:   This module provides a general "estimate" function and EstimationObj
            class for pylogit's logit-type models.
"""
from __future__ import absolute_import

import sys
import time
import numpy as np
import pandas as pd
from scipy.optimize import minimize

from . import choice_calcs as cc
from .choice_calcs import create_matrix_block_indices
from .choice_tools import ensure_PMLE_lambda_is_scalar_or_none
from .choice_tools import ensure_contiguity_in_observation_rows


def ensure_positivity_and_length_of_weights(weights, data):
    assert isinstance(data, pd.DataFrame)
    if weights is None:
        return None
    elif not isinstance(weights, np.ndarray) or weights.ndim != 1:
        msg = '`weights` MUST be a 1D ndarray.'
        raise ValueError(msg)
    elif weights.size != data.shape[0]:
        msg = '`weights` must have the same number of rows as `data`.'
        raise ValueError(msg)
    elif (weights < 0).any():
        msg = '`weights` MUST be >= 0.'
        raise ValueError(msg)
    return None


class EstimationObj(object):
    """
    Generic class for storing pointers to data and methods needed in the
    estimation process. Defines the basic Estimation Object and the various
    methods and attributes that such an object should have. Will be subclassed
    in order to serve the needs of the various types of estimated models.

    Parameters
    ----------
    model_obj : a pylogit.base_multinomial_cm_v2.MNDC_Model instance.
        Should contain the following attributes:

          - alt_IDs
          - choices
          - design
          - intercept_ref_position
          - shape_ref_position
          - utility_transform
    mapping_res : dict.
        Should contain the scipy sparse matrices that map the rows of the long
        format dataframe to various other objects such as the available
        alternatives, the unique observations, etc. The keys that it must have
        are `['rows_to_obs', 'rows_to_alts', 'chosen_row_to_obs']`
    PMLE: None or string value: ['LASSO', 'RIDGE' or 'Tikhonov']
        It determines if a Penalized Maximum Likelihood Estimation should be
        executed. The value of the parameter determines the type of PMLE. The
        None value states that no PMLE is executed. Default = None.
    PMLE_lambda : int, float, long, or None, optional.
        Lambda parameter for LASSO or ridge regression. It should be an int,
        float or long and determines the penalty for the optimization.
        Default = 0.
    zero_vector : 1D ndarray.
        Determines what is viewed as a "null" set of parameters. It is
        explicitly passed because some parameters (e.g. parameters that must be
        greater than zero) have their null values at values other than zero.
    split_params : callable.
        Should take a vector of parameters, `mapping_res['rows_to_alts']`, and
        model_obj.design as arguments. Should return a tuple containing
        separate arrays for the model's shape, outside intercept, and index
        coefficients. For each of these arrays, if this model does not contain
        the particular type of parameter, the callable should place a `None` in
        its place in the tuple.
    constrained_pos : list or None, optional.
        Denotes the positions of the array of estimated parameters that are
        not to change from their initial values. If a list is passed, the
        elements are to be integers where no such integer is greater than
        `init_values.size.` Default == None.
    weights : 1D ndarray or None, optional.
        Allows for the calculation of weighted log-likelihoods. The weights can
        represent various things. In stratified samples, the weights may be
        the proportion of the observations in a given strata for a sample in
        relation to the proportion of observations in that strata in the
        population. In latent class models, the weights may be the probability
        of being a particular class.
    """
    def __init__(self,
                 model_obj,
                 mapping_dict,
                 PMLE,
                 PMLE_lambda,
                 zero_vector,
                 split_params,
                 constrained_pos=None,
                 weights=None):
        # Store pointers to needed objects
        self.alt_id_vector = model_obj.alt_IDs
        self.choice_vector = model_obj.choices
        self.obs_id_vector = model_obj.data[model_obj.obs_id_col].values
        self.design = model_obj.design
        self.intercept_ref_pos = model_obj.intercept_ref_position
        self.shape_ref_pos = model_obj.shape_ref_position

        # Explicitly store pointers to the mapping matrices
        self.rows_to_obs = mapping_dict["rows_to_obs"]
        self.rows_to_alts = mapping_dict["rows_to_alts"]
        self.chosen_row_to_obs = mapping_dict["chosen_row_to_obs"]
        self.rows_to_nests = mapping_dict["rows_to_nests"]
        self.rows_to_mixers = mapping_dict["rows_to_mixers"]

        # Perform necessary checking of PMLE_lambda parameter here!
        ensure_PMLE_lambda_is_scalar_or_none(PMLE_lambda)
        # Ensure the dataset has contiguity in rows with the same obs_id
        ensure_contiguity_in_observation_rows(self.obs_id_vector)
        # Ensure the weights are appropriate for model estimation
        ensure_positivity_and_length_of_weights(weights, model_obj.data)

        # Store the PMLE and PMLE_lambda parameter
        self.PMLE = PMLE
        self.PMLE_lambda = PMLE_lambda

        # Store the constrained parameters
        self.constrained_pos = constrained_pos

        # Store reference to what 'zero vector' is for this model / dataset
        self.zero_vector = zero_vector

        # Store the weights that were passed to the constructor
        self.weights =\
            np.ones(self.design.shape[0]) if weights is None else weights

        # Store the function that separates the various portions of the
        # parameters being estimated (shape parameters, outside intercepts,
        # utility coefficients)
        self.split_params = split_params

        # Store the function that calculates the transformation of the index
        self.utility_transform = model_obj.utility_transform

        # Get the block matrix indices for the hessian matrix.
        self.block_matrix_idxs = create_matrix_block_indices(self.rows_to_obs)

        # Note the following attributes should be set to actual callables that
        # calculate the necessary derivatives in the classes that inherit from
        # EstimationObj
        self.calc_dh_dv = lambda *args: None
        self.calc_dh_d_alpha = lambda *args: None
        self.calc_dh_d_shape = lambda *args: None

        return None

    def convenience_split_params(self, params, return_all_types=False):
        """
        Splits parameter vector into shape, intercept, and index parameters.

        Parameters
        ----------
        params : 1D ndarray.
            The array of parameters being estimated or used in calculations.
        return_all_types : bool, optional.
            Determines whether or not a tuple of 4 elements will be returned
            (with one element for the nest, shape, intercept, and index
            parameters for this model). If False, a tuple of 3 elements will
            be returned with one element for the shape, intercept, and index
            parameters.

        Returns
        -------
        tuple. Will have 4 or 3 elements based on `return_all_types`.
        """
        return self.split_params(params,
                                 self.rows_to_alts,
                                 self.design,
                                 return_all_types=return_all_types)

    def convenience_calc_probs(self, params):
        """
        Calculates the probabilities of the chosen alternative, and the long
        format probabilities for this model and dataset.
        """
        msg = "Method should be defined by descendant classes"
        raise NotImplementedError(msg)

    def convenience_calc_log_likelihood(self, params):
        """
        Calculates the log-likelihood for this model and dataset.
        """
        msg = "Method should be defined by descendant classes"
        raise NotImplementedError(msg)

    def convenience_calc_gradient(self, params):
        """
        Calculates the gradient of the log-likelihood for this model / dataset.
        """
        msg = "Method should be defined by descendant classes"
        raise NotImplementedError(msg)

    def convenience_calc_hessian(self, params):
        """
        Calculates the hessian of the log-likelihood for this model / dataset.
        """
        msg = "Method should be defined by descendant classes"
        raise NotImplementedError(msg)

    def convenience_calc_fisher_approx(self, params):
        """
        Calculates the BHHH approximation of the Fisher Information Matrix for
        this model / dataset.
        """
        msg = "Method should be defined by descendant classes"
        raise NotImplementedError(msg)

    def calc_neg_log_likelihood_and_neg_gradient(self, params):
        """
        Calculates and returns the negative of the log-likelihood and the
        negative of the gradient. This function is used as the objective
        function in scipy.optimize.minimize.
        """
        neg_log_likelihood = -1 * self.convenience_calc_log_likelihood(params)
        neg_gradient = -1 * self.convenience_calc_gradient(params)

        if self.constrained_pos is not None:
            neg_gradient[self.constrained_pos] = 0

        return neg_log_likelihood, neg_gradient

    def calc_neg_hessian(self, params):
        """
        Calculate and return the negative of the hessian for this model and
        dataset.
        """
        return -1 * self.convenience_calc_hessian(params)


class LogitTypeEstimator(EstimationObj):
    """
    Generic class for storing pointers to data and methods needed in the
    estimation process.

    Parameters
    ----------
    model_obj : a pylogit.base_multinomial_cm_v2.MNDC_Model instance.
        Should contain the following attributes:

          - alt_IDs
          - choices
          - design
          - intercept_ref_position
          - shape_ref_position
          - utility_transform
    mapping_res : dict.
        Should contain the scipy sparse matrices that map the rows of the long
        format dataframe to various other objects such as the available
        alternatives, the unique observations, etc. The keys that it must have
        are `['rows_to_obs', 'rows_to_alts', 'chosen_row_to_obs']`
    PMLE: None or string value: ['LASSO', 'RIDGE' or 'Tikhonov']
        It determines if a Penalized Maximum Likelihood Estimation should be
        executed. The value of the parameter determines the type of PMLE. The
        None value states that no PMLE is executed. Default = None.
    PMLE_lambda : int, float, long, or None, optional.
        Lambda parameter for LASSO or ridge regression. It should be an int,
        float or long and determines the penalty for the optimization.
        Default = 0.
    zero_vector : 1D ndarray.
        Determines what is viewed as a "null" set of parameters. It is
        explicitly passed because some parameters (e.g. parameters that must be
        greater than zero) have their null values at values other than zero.
    split_params : callable.
        Should take a vector of parameters, `mapping_res['rows_to_alts']`, and
        model_obj.design as arguments. Should return a tuple containing
        separate arrays for the model's shape, outside intercept, and index
        coefficients. For each of these arrays, if this model does not contain
        the particular type of parameter, the callable should place a `None` in
        its place in the tuple.
    constrained_pos : list or None, optional.
        Denotes the positions of the array of estimated parameters that are
        not to change from their initial values. If a list is passed, the
        elements are to be integers where no such integer is greater than
        `num_params` Default == None.
    weights : 1D ndarray or None, optional.
        Allows for the calculation of weighted log-likelihoods. The weights can
        represent various things. In stratified samples, the weights may be
        the proportion of the observations in a given strata for a sample in
        relation to the proportion of observations in that strata in the
        population. In latent class models, the weights may be the probability
        of being a particular class.

    Attributes
    ----------

    Methods
    -------
    """
    def __init__(self,
                 model_obj,
                 mapping_dict,
                 PMLE,
                 PMLE_lambda,
                 zero_vector,
                 split_params,
                 constrained_pos=None,
                 weights=None):

        kwargs = {"constrained_pos": constrained_pos,
                  "weights": weights}
        super(LogitTypeEstimator, self).__init__(model_obj,
                                                 mapping_dict,
                                                 PMLE,
                                                 PMLE_lambda,
                                                 zero_vector,
                                                 split_params,
                                                 **kwargs)

        return None

    def convenience_calc_probs(self, params):
        """
        Calculates the probabilities of the chosen alternative, and the long
        format probabilities for this model and dataset.
        """
        shapes, intercepts, betas = self.convenience_split_params(params)

        prob_args = [betas,
                     self.design,
                     self.alt_id_vector,
                     self.rows_to_obs,
                     self.rows_to_alts,
                     self.utility_transform]

        prob_kwargs = {"intercept_params": intercepts,
                       "shape_params": shapes,
                       "chosen_row_to_obs": self.chosen_row_to_obs,
                       "return_long_probs": True}
        prob_results = cc.calc_probabilities(*prob_args, **prob_kwargs)

        return prob_results

    def convenience_calc_log_likelihood(self, params):
        """
        Calculates the log-likelihood for this model and dataset.
        """
        shapes, intercepts, betas = self.convenience_split_params(params)

        args = [betas,
                self.design,
                self.alt_id_vector,
                self.rows_to_obs,
                self.rows_to_alts,
                self.choice_vector,
                self.utility_transform]

        kwargs = {"intercept_params": intercepts,
                  "shape_params": shapes,
                  "PMLE": self.PMLE,
                  "PMLE_lambda": self.PMLE_lambda,
                  "weights": self.weights}
        log_likelihood = cc.calc_log_likelihood(*args, **kwargs)

        return log_likelihood

    def convenience_calc_gradient(self, params):
        """
        Calculates the gradient of the log-likelihood for this model / dataset.
        """
        shapes, intercepts, betas = self.convenience_split_params(params)

        args = [betas,
                self.design,
                self.alt_id_vector,
                self.rows_to_obs,
                self.rows_to_alts,
                self.choice_vector,
                self.utility_transform,
                self.calc_dh_d_shape,
                self.calc_dh_dv,
                self.calc_dh_d_alpha,
                intercepts,
                shapes,
                self.PMLE,
                self.PMLE_lambda,
                self.weights]

        return cc.calc_gradient(*args)

    def convenience_calc_hessian(self, params):
        """
        Calculates the hessian of the log-likelihood for this model / dataset.
        """
        shapes, intercepts, betas = self.convenience_split_params(params)

        args = [betas,
                self.design,
                self.alt_id_vector,
                self.rows_to_obs,
                self.rows_to_alts,
                self.utility_transform,
                self.calc_dh_d_shape,
                self.calc_dh_dv,
                self.calc_dh_d_alpha,
                self.block_matrix_idxs,
                intercepts,
                shapes,
                self.PMLE,
                self.PMLE_lambda,
                self.weights]

        return cc.calc_hessian(*args)

    def convenience_calc_fisher_approx(self, params):
        """
        Calculates the BHHH approximation of the Fisher Information Matrix for
        this model / dataset.
        """
        shapes, intercepts, betas = self.convenience_split_params(params)

        args = [betas,
                self.design,
                self.alt_id_vector,
                self.rows_to_obs,
                self.rows_to_alts,
                self.choice_vector,
                self.utility_transform,
                self.calc_dh_d_shape,
                self.calc_dh_dv,
                self.calc_dh_d_alpha,
                intercepts,
                shapes,
                self.PMLE,
                self.PMLE_lambda,
                self.weights]

        return cc.calc_fisher_info_matrix(*args)


def calc_individual_chi_squares(residuals,
                                long_probabilities,
                                rows_to_obs):
    """
    Calculates individual chi-squared values for each choice situation in the
    dataset.

    Parameters
    ----------
    residuals : 1D ndarray.
        The choice vector minus the predicted probability of each alternative
        for each observation.
    long_probabilities : 1D ndarray.
        The probability of each alternative being chosen in each choice
        situation.
    rows_to_obs : 2D scipy sparse array.
        Should map each row of the long format dataferame to the unique
        observations in the dataset.

    Returns
    -------
    ind_chi_squareds : 1D ndarray.
        Will have as many elements as there are columns in `rows_to_obs`. Each
        element will contain the pearson chi-squared value for the given choice
        situation.
    """
    chi_squared_terms = np.square(residuals) / long_probabilities
    return rows_to_obs.T.dot(chi_squared_terms)


def calc_rho_and_rho_bar_squared(final_log_likelihood,
                                 null_log_likelihood,
                                 num_est_parameters):
    """
    Calculates McFadden's rho-squared and rho-bar squared for the given model.

    Parameters
    ----------
    final_log_likelihood : float.
        The final log-likelihood of the model whose rho-squared and rho-bar
        squared are being calculated for.
    null_log_likelihood : float.
        The log-likelihood of the model in question, when all parameters are
        zero or their 'base' values.
    num_est_parameters : int.
        The number of parameters estimated in this model.

    Returns
    -------
    `(rho_squared, rho_bar_squared)` : tuple of floats.
        The rho-squared and rho-bar-squared for the model.
    """
    rho_squared = 1.0 - final_log_likelihood / null_log_likelihood
    rho_bar_squared = 1.0 - ((final_log_likelihood - num_est_parameters) /
                             null_log_likelihood)

    return rho_squared, rho_bar_squared


def calc_and_store_post_estimation_results(results_dict,
                                           estimator):
    """
    Calculates and stores post-estimation results that require the use of the
    systematic utility transformation functions or the various derivative
    functions. Note that this function is only valid for logit-type models.

    Parameters
    ----------
    results_dict : dict.
        This dictionary should be the dictionary returned from
        scipy.optimize.minimize. In particular, it should have the following
        keys: `["fun", "x", "log_likelihood_null"]`.
    estimator : an instance of the EstimationObj class.
        Should contain the following attributes or methods:

          - convenience_split_params
          - convenience_calc_probs
          - convenience_calc_gradient
          - convenience_calc_hessian
          - convenience_calc_fisher_approx
          - choice_vector
          - rows_to_obs

    Returns
    -------
    results_dict : dict.
        The following keys will have been entered into `results_dict`:

          - final_log_likelihood
          - utility_coefs
          - intercept_params
          - shape_params
          - nest_params
          - chosen_probs
          - long_probs
          - residuals
          - ind_chi_squareds
          - rho_squared
          - rho_bar_squared
          - final_gradient
          - final_hessian
          - fisher_info
    """
    # Store the final log-likelihood
    final_log_likelihood = -1 * results_dict["fun"]
    results_dict["final_log_likelihood"] = final_log_likelihood

    # Get the final array of estimated parameters
    final_params = results_dict["x"]

    # Add the estimated parameters to the results dictionary
    split_res = estimator.convenience_split_params(final_params,
                                                   return_all_types=True)
    results_dict["nest_params"] = split_res[0]
    results_dict["shape_params"] = split_res[1]
    results_dict["intercept_params"] = split_res[2]
    results_dict["utility_coefs"] = split_res[3]

    # Get the probability of the chosen alternative and long_form probabilities
    chosen_probs, long_probs = estimator.convenience_calc_probs(final_params)
    results_dict["chosen_probs"] = chosen_probs
    results_dict["long_probs"] = long_probs

    #####
    # Calculate the residuals and individual chi-square values
    #####
    # Calculate the residual vector
    if len(long_probs.shape) == 1:
        residuals = estimator.choice_vector - long_probs
    else:
        residuals = estimator.choice_vector[:, None] - long_probs
    results_dict["residuals"] = residuals

    # Calculate the observation specific chi-squared components
    args = [residuals, long_probs, estimator.rows_to_obs]
    results_dict["ind_chi_squareds"] = calc_individual_chi_squares(*args)

    # Calculate and store the rho-squared and rho-bar-squared
    log_likelihood_null = results_dict["log_likelihood_null"]
    rho_results = calc_rho_and_rho_bar_squared(final_log_likelihood,
                                               log_likelihood_null,
                                               final_params.shape[0])
    results_dict["rho_squared"] = rho_results[0]
    results_dict["rho_bar_squared"] = rho_results[1]

    #####
    # Calculate the gradient, hessian, and BHHH approximation to the fisher
    # info matrix
    #####
    results_dict["final_gradient"] =\
        estimator.convenience_calc_gradient(final_params)
    results_dict["final_hessian"] =\
        estimator.convenience_calc_hessian(final_params)
    results_dict["fisher_info"] =\
        estimator.convenience_calc_fisher_approx(final_params)

    # Store the constrained positions that was used in this estimation process
    results_dict["constrained_pos"] = estimator.constrained_pos

    return results_dict


def estimate(init_values,
             estimator,
             method,
             loss_tol,
             gradient_tol,
             maxiter,
             print_results,
             use_hessian=True,
             just_point=False,
             **kwargs):
    """
    Estimate the given choice model that is defined by `estimator`.

    Parameters
    ----------
    init_vals : 1D ndarray.
        Should contain the initial values to start the optimization process
        with.
    estimator : an instance of the EstimationObj class.
    method : str, optional.
        Should be a valid string for scipy.optimize.minimize. Determines
        the optimization algorithm that is used for this problem.
        Default `== 'bfgs'`.
    loss_tol : float, optional.
        Determines the tolerance on the difference in objective function
        values from one iteration to the next that is needed to determine
        convergence. Default `== 1e-06`.
    gradient_tol : float, optional.
        Determines the tolerance on the difference in gradient values from
        one iteration to the next which is needed to determine convergence.
        Default `== 1e-06`.
    maxiter : int, optional.
        Determines the maximum number of iterations used by the optimizer.
        Default `== 1000`.
    print_res : bool, optional.
        Determines whether the timing and initial and final log likelihood
        results will be printed as they they are determined.
        Default `== True`.
    use_hessian : bool, optional.
        Determines whether the `calc_neg_hessian` method of the `estimator`
        object will be used as the hessian function during the estimation. This
        kwarg is used since some models (such as the Mixed Logit and Nested
        Logit) use a rather crude (i.e. the BHHH) approximation to the Fisher
        Information Matrix, and users may prefer to not use this approximation
        for the hessian during estimation.
    just_point : bool, optional.
        Determines whether or not calculations that are non-critical for
        obtaining the maximum likelihood point estimate will be performed.
        Default == False.

    Return
    ------
    results : dict.
        The dictionary of estimation results that is returned by
        scipy.optimize.minimize. It will also have (at minimum) the following
        keys:
          - "log-likelihood_null"
          - "final_log_likelihood"
          - "utility_coefs"
          - "intercept_params"
          - "shape_params"
          - "nest_params"
          - "chosen_probs"
          - "long_probs"
          - "residuals"
          - "ind_chi_squareds"
          - "rho_squared"
          - "rho_bar_squared"
          - "final_gradient"
          - "final_hessian"
          - "fisher_info"
    """
    if not just_point:
        # Perform preliminary calculations
        log_likelihood_at_zero =\
            estimator.convenience_calc_log_likelihood(estimator.zero_vector)

        initial_log_likelihood =\
            estimator.convenience_calc_log_likelihood(init_values)

        if print_results:
            # Print the log-likelihood at zero
            null_msg = "Log-likelihood at zero: {:,.4f}"
            print(null_msg.format(log_likelihood_at_zero))

            # Print the log-likelihood at the starting values
            init_msg = "Initial Log-likelihood: {:,.4f}"
            print(init_msg.format(initial_log_likelihood))
            sys.stdout.flush()

    # Get the hessian fucntion for this estimation process
    hess_func = None
    if use_hessian and method in ['Newton-CG', 'dogleg', 'trust-ncg',
                                  'trust-krylov', 'trust-exact',
                                  'trust-constr']:
        hess_func = estimator.calc_neg_hessian

    # Estimate the actual parameters of the model
    start_time = time.time()

    results = minimize(estimator.calc_neg_log_likelihood_and_neg_gradient,
                       init_values,
                       method=method,
                       jac=True,
                       hess=hess_func,
                       tol=loss_tol,
                       options={'gtol': gradient_tol,
                                "maxiter": maxiter},
                       **kwargs)

    if not just_point:
        if print_results:
            # Stop timing the estimation process and report the timing results
            end_time = time.time()
            elapsed_sec = (end_time - start_time)
            elapsed_min = elapsed_sec / 60.0
            if elapsed_min > 1.0:
                msg = "Estimation Time for Point Estimation: {:.2f} minutes."
                print(msg.format(elapsed_min))
            else:
                msg = "Estimation Time for Point Estimation: {:.2f} seconds."
                print(msg.format(elapsed_sec))
            print("Final log-likelihood: {:,.4f}".format(-1 * results["fun"]))
            sys.stdout.flush()

        # Store the log-likelihood at zero
        results["log_likelihood_null"] = log_likelihood_at_zero

        # Calculate and store the post-estimation results
        results = calc_and_store_post_estimation_results(results, estimator)

    return results
